#include "tkText.h"
#include "tkTextHighlight.h"

#define SCE_RUBY_DEFAULT 0
#define SCE_RUBY_COMMENTLINE 1
#define SCE_RUBY_NUMBER 2
#define SCE_RUBY_STRING 3
#define SCE_RUBY_CHARACTER 4
#define SCE_RUBY_WORD 5
#define SCE_RUBY_TRIPLE 6
#define SCE_RUBY_TRIPLEDOUBLE 7
#define SCE_RUBY_CLASSNAME 8
#define SCE_RUBY_DEFNAME 9
#define SCE_RUBY_OPERATOR 10
#define SCE_RUBY_IDENTIFIER 11
#define SCE_RUBY_COMMENTBLOCK 12
#define SCE_RUBY_STRINGEOL 13
#define SCE_RUBY_COMMENTDOC 14

static CONST char *styleNames[] = {
    "whitespace", "commentline", "number", "string", "character",
    "word", "triple", "tripledouble", "classname", "defname",
    "operator", "identifier", "commentblock", "stringeol", "commentdoc",
    NULL
};

static void ClassifyWordRb(unsigned int start, unsigned int end, WordList *keywords, char *prevWord) {
    char s[100];
    int len = GetCurrent(s, 100);
    bool wordIsNumber = isdigit(s[0]) != 0;
    if (0 == strcmp(prevWord, "class"))
	ChangeState(SCE_RUBY_CLASSNAME);
    else if (0 == strcmp(prevWord, "module"))
	ChangeState(SCE_RUBY_CLASSNAME);
    else if (0 == strcmp(prevWord, "def"))
	ChangeState(SCE_RUBY_DEFNAME);
    else if (wordIsNumber)
	ChangeState(SCE_RUBY_NUMBER);
    else if (InList(keywords, s, len))
	ChangeState(SCE_RUBY_WORD); // already is SCE_RUBY_WORD...
    // make sure that dot-qualifiers inside the word are lexed correct
    else {
	ChangeState(SCE_RUBY_IDENTIFIER);
#if 0
	for (i = 0; i < len; i++) {
	    if (s[i] == '.') {
		ColourTo(start + i - 1, style);
		ColourTo(start + i, SCE_RUBY_OPERATOR);
	    }
	}
#endif
    }
//	ColourTo(end, style);
    strcpy(prevWord, s);
}

static bool IsRbStringStart(char ch, char chNext, char chNext2) {
    if (ch == '\'' || ch == '"')
	return true;
    if (ch == 'u' || ch == 'U') {
	if (chNext == '"' || chNext == '\'')
	    return true;
	if ((chNext == 'r' || chNext == 'R') && (chNext2 == '"' || chNext2 == '\''))
	    return true;
    }
    if ((ch == 'r' || ch == 'R') && (chNext == '"' || chNext == '\''))
	return true;

    return false;
}

static bool IsRbWordStart(char ch, char chNext, char chNext2) {
    return (iswordchar(ch) && !IsRbStringStart(ch, chNext, chNext2));
}

/* Return the state to use for the string starting at i; *nextIndex will be set to the first index following the quote(s) */
static int GetRbStringState(int i, int *nextIndex) {
    char ch = SafeGetCharAt(i);
    char chNext = SafeGetCharAt(i + 1);

    // Advance beyond r, u, or ur prefix, but bail if there are any unexpected chars
    if (ch == 'r' || ch == 'R') {
	i++;
	ch = SafeGetCharAt(i);
	chNext = SafeGetCharAt(i + 1);
    }
    else if (ch == 'u' || ch == 'U') {
	if (chNext == 'r' || chNext == 'R')
	    i += 2;
	else
	    i += 1;
	ch = SafeGetCharAt(i);
	chNext = SafeGetCharAt(i + 1);
    }

    if (ch != '\"' && ch != '\'') {
	*nextIndex = i + 1;
	return SCE_RUBY_DEFAULT;
    }

    if (ch == chNext && ch == SafeGetCharAt(i + 2)) {
	*nextIndex = i + 3;

	if (ch == '\"')
	    return SCE_RUBY_TRIPLEDOUBLE;
	else
	    return SCE_RUBY_TRIPLE;
    } else {
	*nextIndex = i + 1;

	if (ch == '\"')
	    return SCE_RUBY_STRING;
	else
	    return SCE_RUBY_CHARACTER;
    }
}

static int ColouriseRuby(LexerArgs *args)
{
    WordList *keywords = args->sharedPtr->wordListPtrs[0];
    char prevWord[200];
    int nextIndex = 0;
    char chNext2;

    BeginStyling(args, args->firstLine, args->lastLine);
#if 0
    if (vars.lineLength == 0)
	return 0;
#endif
    prevWord[0] = '\0';

    for (; More(); Forward()) {

	chNext2 = SafeGetCharAt(vars.charIndex + 2);
#if 0
	if (vars.atLineEnd) {
	    if ((vars.style == SCE_RUBY_DEFAULT) || (vars.style == SCE_RUBY_TRIPLE) || (vars.style == SCE_RUBY_TRIPLEDOUBLE)) {
		// Perform colourisation of white space and triple quoted strings at end of each line to allow
		// tab marking to work inside white space and triple quoted strings
		Forward();
	    }
	}
#endif

	// Determine if the current state should terminate.
	if (vars.style == SCE_RUBY_OPERATOR) {
	   SetStyle(SCE_RUBY_DEFAULT);
	} else if (vars.style == SCE_RUBY_WORD) {
	    if (!iswordchar(vars.chCur)) {
		ClassifyWordRb(vars.startIndex, vars.charIndex - 1, keywords, prevWord);
		SetStyle(SCE_RUBY_DEFAULT);
	    }
	} else if (vars.style == SCE_RUBY_COMMENTLINE || vars.style == SCE_RUBY_COMMENTBLOCK) {
	    if (vars.atLineEnd) {
		SetStyle(SCE_RUBY_DEFAULT);
	    }
	} else if (vars.style == SCE_RUBY_STRING) {
	    if (vars.chCur == '\\') {
		if (vars.chNext == '\"' || vars.chNext == '\'' || vars.chNext == '\\') {
		    Forward();
		}
	    } else if (vars.chCur == '\"') {
		ForwardSetStyle(SCE_RUBY_DEFAULT);
	    } else if (vars.atLineEnd) {
		ChangeState(SCE_RUBY_STRINGEOL);
		ForwardSetStyle(SCE_RUBY_DEFAULT);
	    }
	} else if (vars.style == SCE_RUBY_CHARACTER) {
	    if (vars.chCur == '\\') {
		if (vars.chNext == '\"' || vars.chNext == '\'' || vars.chNext == '\\') {
		    Forward();
		}
	    } else if (vars.chCur == '\'') {
		ForwardSetStyle(SCE_RUBY_DEFAULT);
	    } else if (vars.atLineEnd) {
		ChangeState(SCE_RUBY_STRINGEOL);
		ForwardSetStyle(SCE_RUBY_DEFAULT);
	    }
	} else if (vars.style == SCE_RUBY_TRIPLE) {
	    if (vars.chCur == '\'' && vars.chNext == '\'') {
		Forward();
		Forward();
		if (vars.chCur == '\'') {
		    ForwardSetStyle(SCE_RUBY_DEFAULT);
		}
	    }
	} else if (vars.style == SCE_RUBY_COMMENTDOC) {
	    // =end terminates the comment block
	    if (MatchStr("=end")) {
		Forward();
		Forward();
		Forward();
		ForwardSetStyle(SCE_RUBY_DEFAULT);
	    }
	}

	// Determine if a new state should be entered.
	if (vars.style == SCE_RUBY_DEFAULT) {
	    if (IsRbWordStart(vars.chCur, vars.chNext, chNext2)) {
		SetStyle(SCE_RUBY_WORD);
	    } else if (vars.chCur == '#') {
		SetStyle(vars.chNext == '#' ? SCE_RUBY_COMMENTBLOCK : SCE_RUBY_COMMENTLINE);
	    } else if (vars.chCur == '=' && vars.chNext == 'b') {
		// =begin indicates the start of a comment (doc) block
		if(MatchStr("=begin")) {
		   SetStyle(SCE_RUBY_COMMENTDOC);
		}
	    } else if (IsRbStringStart(vars.chCur, vars.chNext, chNext2)) {
		SetStyle(GetRbStringState(vars.charIndex, &nextIndex));
		if (nextIndex != vars.charIndex + 1) {
		    do {
			Forward();
		    } while (vars.charIndex < nextIndex - 1);
		}
	    } else if (isoperator(vars.chCur)) {
		SetStyle(SCE_RUBY_OPERATOR);
	    }
	}
    }
    if (vars.style == SCE_RUBY_WORD) {
	ClassifyWordRb(vars.startIndex, vars.lineLength-1, keywords, prevWord);
    }
    Complete();
    return 0;
}

Lexer lexerRuby = { "ruby", styleNames, ColouriseRuby, (LexerFunction) NULL };

